/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2019  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * format.c -- print nicely formatted text
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>

#include "safe.h"
#include "compat.h"
#include "format.h"

/* Get the width of a TAB character at column `col'. */
#define get_tab_width(col) (format_tab_stop - ((col) % format_tab_stop))

struct string_list {
  char *str;
  struct string_list *next;
};

/*
 * Static helper functions
 */
static int format_list(struct string_list *, unsigned short, unsigned short *,
			unsigned short, unsigned short, const char *);

/* The number of spaces per tab. */
unsigned short format_tab_stop = FORMAT_TAB_STOP;

/*
 * These simply call the corresponding v* functions (see below).
 */

int format(FILE *stream, unsigned short width, unsigned short *col,
	   unsigned short indent, unsigned short subindent,
	   const char *fmt, ...) {
  va_list ap;
  int ret;

  va_start(ap, fmt);
  ret = vformat(stream, width, col, indent, subindent, fmt, ap);
  va_end(ap);

  return ret;
}

char *format_string(unsigned short width, unsigned short *col,
		    unsigned short indent, unsigned short subindent,
		    const char *fmt, ...) {
  va_list ap;
  char *ret;

  va_start(ap, fmt);
  ret = vformat_string(width, col, indent, subindent, fmt, ap);
  va_end(ap);

  return ret;
}

/*
 * Print a formatted string to `stream'. See format_plain().
 */
int vformat(FILE *stream, unsigned short width, unsigned short *col,
	    unsigned short indent, unsigned short subindent,
	    const char *fmt, va_list ap) {
  char *str;
  struct string_list *string_list;
  int ret;

  ret = vasprintf(&str, fmt, ap);
  if (ret < 0)
    return ret;

  string_list = malloc(sizeof(*string_list));
  if (!string_list) {
    free(str);
    return ret;
  }

  ret = format_list(string_list, width, col, indent, subindent, str);

  free(str);

  if (ret) {
    free(string_list);
    return ret;
  }

  /* Print the formatted string. */
  while (string_list) {
    struct string_list *old = string_list;
    fputs(string_list->str, stream);
    string_list = string_list->next;
    free(old->str);
    free(old);
  }

  return ret;
}

/*
 * Like vformat(), but returns a string instead (or NULL on
 * error). See vformat() and format_list().
 */
char *vformat_string(unsigned short width, unsigned short *col,
		     unsigned short indent, unsigned short subindent,
		     const char *fmt, va_list ap) {
  char *str;
  size_t size, index;
  struct string_list *string_list, *pos;
  int ret, errno_save;

  ret = vasprintf(&str, fmt, ap);
  if (ret < 0)
    return NULL;

  string_list = malloc(sizeof(*string_list));
  if (!string_list) {
    free(str);
    return NULL;
  }

  ret = format_list(string_list, width, col, indent, subindent, str);

  free(str);

  if (ret) {
    free(string_list);
    return NULL;
  }

  /* Get the size of the string. */
  size = 0;
  for (pos = string_list; pos; pos = pos->next)
    size += strlen(pos->str);
  size++;			/* For '\0' */

  /* Construct the string. */
  index = 0;
  str = malloc(sizeof(*str) * size);
  errno_save = errno;		/* free() might set `errno' */
  while (string_list) {
    struct string_list *old = string_list;

    if (str) {
      size_t len = strlen(string_list->str);
      memcpy(str + index, string_list->str, len);
      index += len;
    }

    string_list = string_list->next;
    free(old->str);
    free(old);
  }
  errno = errno_save;
  assert(index == size);
  str[index] = '\0';

  return str;
}

/***************************\
|* Static helper functions *|
\***************************/

/*
 * Format a string, `str' and return it as a list of strings in
 * `formatted' (which should be concatenated), while staying in the
 * constraints of `width' (or no constraints if `width' is 0). '*col'
 * should be initialized to zero and is used to keep track of the
 * current column. `indent' will be used to indent the current line of
 * text, and `subindent' is the indent to use on following
 * lines. Returns zero on success or nonzero on error, in which case
 * `errno' will be set to indicate the error. If nonzero is returned,
 * `formatted' is guaranteed to be empty.
 *
 * This algorithm only makes a half-hearted attempt to preserve manual
 * indentation. That part could be improved.
 */
static int format_list(struct string_list *formatted, unsigned short width,
		       unsigned short *col, unsigned short indent,
		       unsigned short subindent, const char *str) {
  const char *pos;
  size_t size, index;
  /* The number of spaces required to reach `indent'. */
  unsigned short n_indent;
  enum { SPLIT_NONE = 0, SPLIT_SPACE, SPLIT_WORD } split;
  int ret;

  if (width) {
    assert(indent < width);
    assert(*col < width);
  }

  formatted->next = NULL;

  if (!*str) {
    formatted->str = strdup("");
    return !formatted->str;
  }

  if (*str == '\n') {
    /* Add the newlines. */
    for (size = 0; *str == '\n'; str++)
      size++;
    formatted->str = malloc(sizeof(*formatted->str) * (size + 1));
    if (!formatted->str)
      return 1;
    memset(formatted->str, '\n', size);
    formatted->str[size] = '\0';

    formatted->next = malloc(sizeof(*formatted->next));
    if (!formatted->next) {
      free(formatted->str);
      return 1;
    }

    *col = 0;

    /* Format the rest of the string. Note: `subindent' is repeated on
       purpose, because that should be the indentation for subsequent
       lines. */
    ret = format_list(formatted->next, width, col, subindent, subindent, str);
    if (ret) {
      free(formatted->str);
      free(formatted->next);
    }

    return ret;
  }

  /* Get the size of the current string. */
  split = SPLIT_NONE;
  size = n_indent = (*col < indent) ? indent - *col : 0;
  for (pos = str; *pos && *pos != '\n'; pos++) {
    unsigned short char_width;

    /* We don't handle non-printable characters except TAB (which is
       considered non-printable). */
    if (*pos != '\t' && !isprint(*pos)) {
      errno = EINVAL;
      return 1;
    }

    if (*pos == '\t')
      char_width = get_tab_width(*col + size);
    else
      char_width = 1;

    if (width && size + char_width > width - *col) {
      split = isspace(*pos) ? SPLIT_SPACE : SPLIT_WORD;
      break;
    }

    size += char_width;
  }

  /* Allocate the string. '+ !!split' because we will add a newline if
     `split'. */
  formatted->str = malloc(sizeof(*formatted->str) * (size + 1 + !!split));
  if (!formatted->str)
    return 1;

  /* Indent with spaces. */
  memset(formatted->str, ' ', n_indent);

  /* Copy over the unformatted string, converting tabs to spaces. */
  index = n_indent;
  for (pos = str; *pos && *pos != '\n'; pos++) {
    if (*pos == '\t') {
      unsigned short tab_width = get_tab_width(*col + index);

      /* Make sure we don't exceed the `width' constraint. */
      if (index + tab_width > size) {
	assert(width);
	assert(index + tab_width > width - *col);
	break;
      }

      memset(formatted->str + index, ' ', tab_width);
      index += tab_width;
    }
    else {
      /* Make sure we don't exceed the `width' constraint. */
      if (index + 1 > size) {
	assert(width);
	assert(index + 1 > width - *col);
	break;
      }

      formatted->str[index++] = *pos;
    }
  }
  assert(index == size);
  formatted->str[index] = '\0';

  if (split == SPLIT_WORD) {
    size_t i, j;

    /* We split a word, so put that word on the next line instead. */
    for (i = size - 1; i && !isspace(formatted->str[i]); i--);
    for (j = i; j && isspace(formatted->str[j]); j--);

    if (j) {
      do pos--; while (pos > str && !isspace(*pos));
      if (isspace(*pos)) pos++;
      size = j + 1;
    }
    else {
      /* The word takes up the whole line, so just put a dash at the
	 end if we have room and continue it on the next line. */
      if (size > 1) {
	formatted->str[size - 1] = '-';
	pos--;
      }
    }
  }
  else if (split == SPLIT_SPACE) {
    /* We split in the middle of some space. Get rid of it. */
    do
      size--;
    while (size && isspace(formatted->str[size]));
    if (!isspace(formatted->str[size]))
      size++;
    while (*pos && isspace(*pos))
      pos++;
  }

  if (split) {
    formatted->str[size++] = '\n';
    formatted->str[size] = '\0';
    *col = 0;
  }
  else {
    *col += size;
    if (width)
      assert(*col <= width);
  }

  formatted->next = malloc(sizeof(*formatted->next));
  if (!formatted->next) {
    free(formatted->str);
    return 1;
  }

  /* Format the rest of the string. Note: `subindent' is repeated on
     purpose, because that should be the indentation for subsequent
     lines. */
  ret = format_list(formatted->next, width, col, subindent, subindent, pos);
  if (ret) {
    free(formatted->str);
    free(formatted->next);
  }

  return ret;
}
