/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2019  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * options.c -- routines for parsing options
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <dirent.h>
#include <limits.h>
#include <errno.h>
#include <assert.h>

#include "safe.h"
#include "io.h"
#include "format.h"
#include "options.h"

#define OPT_END(opt) ((opt)->short_opt == '\0' && (opt)->long_opt == NULL)

/* The argument separator for long options. */
#define ARG_SEPARATOR '='

/* The absolute maximum line length for usage messages. Undefine for
   no max. */
#define USAGE_MAX_WIDTH		80
#define USAGE_DEFAULT_WIDTH	USAGE_MAX_WIDTH

typedef union opt_arg {
  int arg_bool;
  long arg_int;
  double arg_float;
  const char *arg_string;
  FILE *arg_file;
  DIR *arg_directory;
} OPT_ARG;

/* This is for formatting the usage message. */
struct help_callback_data {
  const char *prog_name, *usage, *desc, *notes;
  OPTION *options;
};

/*
 * Static helper functions
 */
static int parse_short_opt(const char *, const char *,
			   const char *, const OPTION *, int);
static int parse_long_opt(const char *, const char *,
			  const char *, const OPTION *, int);
static int handle_option(const char *, int, const OPTION *, const char *);
static int find_unambiguously(const OPTION *, const char *);
static int free_opt(const char *, const OPTION *);
static int free_opts(const char *, const OPTION *);
static int format_help_callback(void *, char *);

/*
 * Parse options. Returns zero on success or nonzero on error, in
 * which case an error message will be printed to `stderr'. The last
 * element of `options' must have a `short_opt' of '\0' and a
 * `long_opt' of NULL.
 */
int parse_opts(int *p_argc, char **argv, const char *usage, const char *desc,
	       const char *notes, const OPTION *options, int flags) {
  int argc = *p_argc;
  int ret = 0;
  OPTION *actual_opts;

  if (flags & OPT_PERMUTE) {
    /* This will be set to the number of positional arguments (plus
       'argv[0]') when we're done. */
    *p_argc = 1;
  }

  if (flags & OPT_HELP) {
    size_t n_opts;
    OPTION *opt;
    struct help_callback_data data;

    /* Count the options and make sure there is no option conflicting
       with our `-h' or `--help'. */
    for (n_opts = 0; !OPT_END(options + n_opts); n_opts++) {
      const OPTION *opt = options + n_opts;
      if (flags & OPT_HELP_SHORT)
	assert(opt->short_opt != 'h');
      if (flags & OPT_HELP_LONG)
	assert(!opt->long_opt || strcmp(opt->long_opt, "help"));
    }

    /* Create a new list of options and copy these ones over. */
    actual_opts = xmalloc(sizeof(*actual_opts) * (n_opts + 2));
    memcpy(actual_opts, options, sizeof(*actual_opts) * n_opts);

    opt = actual_opts + n_opts;
    /* *2 to initialize the help option and also set the last one to
       *all zeros. */
    memset(opt, 0, sizeof(*actual_opts) * 2);

    /* Add the help option. */
    data.prog_name	= argv[0];
    data.usage		= usage;
    data.desc		= desc;
    data.notes		= notes;
    data.options	= actual_opts;
    if (flags & OPT_HELP_SHORT)
      opt->short_opt	= 'h';
    if (flags & OPT_HELP_LONG)
      opt->long_opt	= "help";
    opt->has_arg	= OPT_NONE;
    opt->callback_none	= format_help_callback;
    opt->data		= &data;
    opt->help		= "print this help and exit";
  }
  else {
    /* This is slightly evil (casting it to a non-const), but we know
       we are not going to modify it now. */
    actual_opts = (OPTION *)options;
  }

  /* Initalize all the `found_opt', `found_arg', `arg', and `argstr'
     values to zero. */
  for (const OPTION *opt = actual_opts; !OPT_END(opt); opt++) {
    if (opt->found_opt)
      *(opt->found_opt) = 0;
    if (opt->found_arg)
      *(opt->found_arg) = 0;

    if (opt->has_arg) {
      OPT_ARG *arg = opt->arg;

      if (arg) {
	switch (opt->type) {
	case OPT_BOOL:
	  arg->arg_bool = 0;
	  break;
	case OPT_INT:
	  arg->arg_int = 0;
	  break;
	case OPT_FLOAT:
	  arg->arg_float = 0;
	  break;
	case OPT_STRING:
	  arg->arg_string = NULL;
	  break;
	case OPT_FILE:
	  arg->arg_file = NULL;
	  break;
	case OPT_DIRECTORY:
	  arg->arg_directory = NULL;
	  break;
	default:
	  assert(!"Invalid option type");
	}
      }

      if (opt->argstr)
	*(opt->argstr) = NULL;
    }
  }

  for (int i = 1; i < argc; i++) {
    char *optstr = argv[i];
    const char *arg = (i + 1 < argc) ? argv[i+1] : NULL;

    /* Stop if we see "--". */
    if (!strcmp(optstr, "--"))
      break;

    /* Try to parse it as a short option. */
    ret = parse_short_opt(argv[0], optstr, arg, actual_opts, flags);
    if (ret == 0 || ret == 1) {
      if (ret)
	i++;
      continue;
    }
    if (ret == -1) {
      free_opts(argv[0], actual_opts);
      break;
    }

    /* Didn't work? Try to parse it as a long option. */
    ret = parse_long_opt(argv[0], optstr, arg, actual_opts, flags);
    if (ret == 0 || ret == 1) {
      if (ret)
	i++;
      continue;
    }
    if (ret == -1) {
      free_opts(argv[0], actual_opts);
      break;
    }

    /* Still didn't work? It must not be an option. */
    if (flags & OPT_PERMUTE) {
      assert(!(flags & OPT_STOP_AFTER_ARG));
      argv[(*p_argc)++] = optstr;
    }
    else if (flags & OPT_STOP_AFTER_ARG) {
      break;
    }
  }

  if (flags & OPT_HELP)
    free(actual_opts);

  /* Make sure `argv' ends in NULL. */
  if (*p_argc < argc)
    argv[*p_argc] = NULL;

  return ret == -1;
}

/*
 * Format help output from `usage' and `options', printing it to
 * `file'. Returns zero on success, nonzero on error in which case
 * `errno' will be set to indicate the error.
 */
int format_help(FILE *file, const char *prog_name, const char *usage,
		const char *desc, const char *notes, const OPTION *options) {
  unsigned short width, cur_col = 0;
  int long_required_arg, short_opts;

  /* Get the terminal width. */
#ifdef TIOCGWINSZ
  struct winsize ws;
  if (ioctl(fileno(file), TIOCGWINSZ, &ws) >= 0)
    width = ws.ws_col;
  else
#endif
    width = USAGE_DEFAULT_WIDTH;
#ifdef USAGE_MAX_WIDTH
  if (width > USAGE_MAX_WIDTH)
    width = USAGE_MAX_WIDTH;
#endif

  if (format(file, width, &cur_col, 0, 0, "Usage: %s %s\n%s\n%s",
	     prog_name, usage, desc, OPT_END(options) ? "" : "\n"))
    return 1;

  /*
   * Check whether there are long options with requried arguments that
   * have equivalent short options and whether there are short options
   * at all.
   */
  long_required_arg = short_opts = 0;
  for (const OPTION *opt = options; !OPT_END(opt); opt++) {
    if (opt->short_opt) {
      short_opts = 1;
      if (opt->long_opt && opt->has_arg == OPT_REQUIRED) {
	long_required_arg = 1;
	break;
      }
    }
  }

  if (long_required_arg &&
      format(file, width, &cur_col, 0, 0, "Mandatory arguments to long "
	     "options are mandatory for short options too.\n", file))
    return 1;

  /* Print the help for each option. */
  for (const OPTION *opt = options; !OPT_END(opt); opt++) {
    if (opt->short_opt) {
      if (format(file, width, &cur_col, 0, 0, "  -%c", opt->short_opt))
	return 1;
    }
    else if (short_opts) {
      if (format(file, width, &cur_col, 0, 0, "    "))
	return 1;
    }

    if (opt->long_opt) {
      if (format(file, width, &cur_col, 0, 0, "%c --%s",
		  opt->short_opt ? ',' : ' ', opt->long_opt))
	return 1;
    }

    if (opt->help && format(file, width, &cur_col, 30, 32, "%s", opt->help))
      return 1;

    if (format(file, width, &cur_col, 0, 0, "\n"))
      return 1;
  }

  if (notes && format(file, width, &cur_col, 0, 0, "\n%s\n", notes))
    return 1;

  return 0;
}

/***************************\
|* Static helper functions *|
\***************************/

/*
 * Parse a short option, returning 0 on sucess if `argument' wasn't
 * used, 1 on success if `argument' was used, 2 if it's not a short
 * option, and -1 if a fatal error occured (in which case an error
 * message will be printed to `stderr').
 */
static int parse_short_opt(const char *prog_name, const char *optstr,
			   const char *argument, const OPTION *options,
			   int flags) {
  int used_arg = 0;

  if (*(optstr++) != '-' || *optstr == '-' || !*optstr)
    return 2;

  for (char c = *optstr; (c = *optstr); optstr++) {
    const OPTION *opt;
    const char *argstr = NULL;

    /* Find the short option. */
    for (opt = options; !OPT_END(opt); opt++) {
      if (opt->short_opt == c)
	break;
    }

    if (OPT_END(opt)) {
      fprintf(stderr, "%s: '%c': invalid option\n", prog_name, c);
      return -1;
    }

    /* Get the argument if there is one. */
    if (optstr[1]) {
      if (opt->has_arg)
	argstr = optstr + 1;
      else if (!(flags & OPT_BUNDLE))
	return 2;
    }
    else if (opt->has_arg == OPT_REQUIRED) {
      argstr = argument;
      used_arg = 1;
    }

    if (handle_option(prog_name, 1, opt, argstr))
      return -1;

    if (!(flags & OPT_BUNDLE) || opt->has_arg)
      break;
  }

  return used_arg;
}

/*
 * Parse a long option, returning 0 on sucess if `argument' wasn't
 * used, 1 on success if `argument' was used, 2 if it's not a long
 * option, and -1 if a fatal error occured (in which case an error
 * message will be printed to `stderr').
 */
static int parse_long_opt(const char *prog_name, const char *optstr,
			  const char *argument, const OPTION *options,
			  int flags) {
  int used_arg = 0;
  int index;
  OPTION opt;
  const char *argstr;

  if (*(optstr++) != '-')
    return 2;

  /* If we're not bundling options, long options can start with a
     single '-' too. */
  if (flags & OPT_BUNDLE && *(optstr++) != '-')
    return 2;

  if (!*optstr)
    return 2;

  /* An option cannot start with ARG_SEPARATOR. */
  if (*optstr == ARG_SEPARATOR) {
    fprintf(stderr, "%s: %s: invalid option\n", prog_name, optstr);
    return -1;
  }

  /* Search for the option. */
  index = find_unambiguously(options, optstr);
  if (index < 0) {
    fprintf(stderr, "%s: %s: invalid option\n", prog_name, optstr);
    return -1;
  }

  opt = options[index];

  argstr = strchr(optstr, ARG_SEPARATOR);
  if (argstr) {
    argstr++;
  }
  else if (opt.has_arg == OPT_REQUIRED) {
    argstr = argument;
    used_arg = 1;
  }

  return handle_option(prog_name, 0, &opt, argstr) ? -1 : used_arg;
}

/*
 * Handle an option. Returns zero on success, nonzero on error (in
 * which case an error message will be printed to `stderr').
 */
static int handle_option(const char *prog_name, int is_short,
			 const OPTION *opt, const char *argstr) {
  OPT_ARG arg;
  OPT_ARG *p_arg = opt->arg ? opt->arg : &arg;
  int ret = 0;

  if (opt->has_arg == OPT_REQUIRED && !argstr) {
    char short_opt[2];
    const char *optstr;

    if (is_short) {
      short_opt[0] = opt->short_opt;
      short_opt[1] = '\0';
      optstr = short_opt;
    }
    else
      optstr = opt->long_opt;

    fprintf(stderr, "%s: '%s': option requires argument\n", prog_name, optstr);

    return 1;
  }

  /*
   * Process the argument.
   */
  if (argstr) {
    if (opt->has_arg == OPT_NONE) {
      if (is_short)
	fprintf(stderr, "%s: '%c': option does not take argument\n",
		prog_name, opt->short_opt);
      else
	fprintf(stderr, "%s: %s: option does not take argument\n",
		prog_name, opt->long_opt);

      return 1;
    }

    /* We are going to overwrite the previous argument. */
    if (free_opt(prog_name, opt))
      return 1;

    switch (opt->type) {
      size_t arglen;
      long arg_int;
      char *endptr;

    case OPT_BOOL:
      arglen = strlen(argstr);
      if (!strncasecmp(argstr, "yes", arglen) ||
	  !strncasecmp(argstr, "true", arglen)) {
	p_arg->arg_bool = 1;
	break;
      }
      if (!strncasecmp(argstr, "no", arglen) ||
	  !strncasecmp(argstr, "false", arglen)) {
	p_arg->arg_bool = 0;
	break;
      }

      /* fall-through */

    case OPT_INT:
      errno = 0;
      arg_int = strtol(argstr, &endptr, 0);

      if (*endptr) {
	fprintf(stderr, "%s: %s: not a%s\n", prog_name, argstr,
		(opt->type == OPT_BOOL) ? " yes or no value" : "n integer");
	return 1;
      }

      if (opt->type == OPT_INT) {
	if (!errno && (arg_int < opt->ibound.lower ||
		       arg_int > opt->ibound.upper))
	  errno = ERANGE;

	if (errno) {
	  fprintf(stderr, "%s: %s: %s\n", prog_name, argstr, strerror(errno));
	  return 1;
	}
      }

      if (opt->type == OPT_BOOL)
	p_arg->arg_bool = !!arg_int;
      else
	p_arg->arg_int = arg_int;

      break;

    case OPT_FLOAT:
      errno = 0;
      p_arg->arg_float = strtod(argstr, &endptr);

      if (*endptr) {
	fprintf(stderr, "%s: %s: not a floating-point number\n",
		prog_name, argstr);
	return 1;
      }

      if (!errno && (p_arg->arg_float < opt->fbound.lower ||
		     p_arg->arg_float > opt->fbound.upper))
	errno = ERANGE;

      if (errno) {
	fprintf(stderr, "%s: %s: %s\n", prog_name, argstr, strerror(errno));
	return 1;
      }

      break;

    case OPT_STRING:
      p_arg->arg_string = argstr;
      break;

    case OPT_FILE:
      p_arg->arg_file = file_open(argstr, opt->file_mode);
      if (!p_arg->arg_file) {
	fprintf(stderr, "%s: cannot open %s: %s\n",
		prog_name, argstr, strerror(errno));
	return 1;
      }

      if (p_arg->arg_file == stdin)
	argstr = "<stdin>";
      else if (p_arg->arg_file == stdout)
	argstr = "<stdout>";

      break;

    case OPT_DIRECTORY:
      p_arg->arg_directory = opendir(argstr);
      if (!p_arg->arg_directory) {
	fprintf(stderr, "%s: cannot open directory %s: %s\n",
		prog_name, argstr, strerror(errno));
	return 1;
      }

      break;

    default:
      assert(!"Invalid option type");
    }
  }

  if (opt->found_opt)
    *(opt->found_opt) = 1;
  if (opt->found_arg)
    *(opt->found_arg) = !!argstr;

  if (opt->argstr)
    *(opt->argstr) = argstr;

  /*
   * Call the callback.
   */
  if (opt->callback_none) {
    /* Initialize to zeros in case the callback forgets to add a
       terminating null byte. */
    char err[OPT_ERR_MAX] = { 0 };

    if (opt->has_arg == OPT_NONE) {
      ret = opt->callback_none(opt->data, err);
    }
    else {
      switch (opt->type) {
      case OPT_BOOL:
	ret = opt->callback_bool(!!argstr, p_arg->arg_bool, opt->data, err);
	break;
      case OPT_INT:
	ret = opt->callback_int(!!argstr, p_arg->arg_int, opt->data, err);
	break;
      case OPT_FLOAT:
	ret = opt->callback_float(!!argstr, p_arg->arg_float, opt->data, err);
	break;
      case OPT_STRING:
	ret =
	  opt->callback_string(!!argstr, p_arg->arg_string, opt->data, err);
	break;
      case OPT_FILE:
	ret = opt->callback_file(!!argstr, argstr, p_arg->arg_file,
				 opt->data, err);
	break;
      case OPT_DIRECTORY:
	ret = opt->callback_directory(!!argstr, argstr, p_arg->arg_directory,
				      opt->data, err);
	break;
      default:
	assert(!"Invalid option type");
      }
    }

    if (ret)
      fprintf(stderr, "%s: %s\n", prog_name, err);
  }

  /*
   * If we're not going to return the file/directory we opened, we
   * should close it now.
   */
  if (!opt->arg) {
    switch (opt->type) {
    case OPT_FILE:
      if (file_close(p_arg->arg_file)) {
	fprintf(stderr, "%s: cannot close %s: %s\n",
		prog_name, argstr, strerror(errno));
	return 1;
      }
      break;

    case OPT_DIRECTORY:
      if (closedir(p_arg->arg_directory)) {
	fprintf(stderr, "%s: cannot close directory %s: %s\n",
		prog_name, argstr, strerror(errno));
	return 1;
      }
      break;

    default:
      /* Don't warn on -Wswitch */
      break;
    }
  }

  return ret;
}

/*
 * Find the long option `needle' in `haystack', allowing abbreviations
 * as long as they are not ambiguous. Returns the index of the found
 * string in `haystack' if found, or -1 if it could not be found
 * unambiguously.
 */
static int find_unambiguously(const OPTION *haystack, const char *needle) {
  int n_haystack, n_matches, index;
  struct {
    const char *match;
    int index;
  } *matches;
  size_t needle_size;
  char *equals;
  int ret;

  /* Get the size of the matches and the haystack. */
  n_matches = 0;
  for (n_haystack = 0; !OPT_END(haystack + n_haystack); n_haystack++) {
    if (haystack[n_haystack].long_opt)
      n_matches++;
  }

  /* Construct the matches. */
  matches = xmalloc(sizeof(*matches) * n_matches);
  index = 0;
  for (int i = 0; i < n_haystack; i++) {
    const char *str = haystack[i].long_opt;
    if (str) {
      assert(index < n_matches);
      matches[index].match = str;
      matches[index].index = i;
      index++;
    }
  }
  assert(index == n_matches);

  /* We stop parsing at ARG_SEPARATOR. */
  equals = strchr(needle, ARG_SEPARATOR);
  if (equals)
    needle_size = equals - needle;
  else
    needle_size = strlen(needle);

  /* Search for the needle. */
  for (size_t i = 0; i < needle_size; i++) {
    /* Remove all non-matching items. */
    for (int j = 0; j < n_matches; j++) {
      if (matches[j].match[i] != needle[i])
	matches[j] = matches[--n_matches];
    }

    /* There's no point in searching any more if all matches are
       gone. */
    if (!n_matches)
      break;
  }

  if (n_matches == 1)
    ret = matches->index;
  else
    ret = -1;

  free(matches);
  return ret;
}

/*
 * Release resources (currently just opened files/directories)
 * associated with `opt'. Returns zero on success or 1 on error, in
 * which case an error message will be printed to `stderr'.
 */
static int free_opt(const char *prog_name, const OPTION *opt) {
  OPT_ARG *arg = opt->arg;

  if (!arg)
    return 0;

  switch (opt->type) {
  case OPT_BOOL:
    arg->arg_bool = 0;
    break;
  case OPT_INT:
    arg->arg_int = 0;
    break;
  case OPT_FLOAT:
    arg->arg_float = 0;
    break;
  case OPT_STRING:
    arg->arg_string = NULL;
    break;

  case OPT_FILE:
    if (arg->arg_file) {
      if (file_close(arg->arg_file)) {
	fprintf(stderr, "%s: cannot close %s: %s\n", prog_name,
		opt->argstr ? *(opt->argstr) : "file", strerror(errno));
	return 1;
      }
      arg->arg_file = NULL;
    }
    break;

  case OPT_DIRECTORY:
    if (arg->arg_directory) {
      if (closedir(arg->arg_directory)) {
	fprintf(stderr, "%s: cannot close directory%s%s: %s\n", prog_name,
		opt->argstr ? " " : "", opt->argstr ? *(opt->argstr) : "",
		strerror(errno));
	return 1;
      }
      arg->arg_directory = NULL;
    }

    break;

  default:
    assert(!"Invalid option type");
  }

  if (opt->argstr)
    *(opt->argstr) = NULL;

  return 0;
}

/*
 * Call free_opt() for each option in `options'. Returns the number of
 * failures.
 */
static int free_opts(const char *prog_name, const OPTION *options) {
  int ret = 0;

  for (const OPTION *opt = options; !OPT_END(opt); opt++)
    ret += free_opt(prog_name, opt);

  return ret;
}

/*
 * Format and print a help message to `stdout'. Does not return on
 * success.
 */
static int format_help_callback(void *data, char *err) {
  struct help_callback_data *d = data;
  int ret = format_help(stdout, d->prog_name, d->usage,
			d->desc, d->notes, d->options);

  if (ret) {
    /* So the error message doesn't mingle with what we've already
       printed. */
    fflush(stdout);
    snprintf(err, OPT_ERR_MAX, "cannot print help message: %s",
	     strerror(errno));
    return ret;
  }

  /* We don't really need to do this since we're about to exit(), but
     it makes valgrind happy. */
  free(d->options);

  exit(0);
}
