/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2019  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * options.h -- routines for parsing options
 */

#ifndef _OPTIONS_H
#define _OPTIONS_H

#include <stdio.h>
#include <dirent.h>

/* The maximum length of an error message. */
#define OPT_ERR_MAX 256

#define OPT_PERMUTE		0b00000001
#define OPT_STOP_AFTER_ARG	0b00000010
#define OPT_BUNDLE		0b00000100
#define OPT_HELP_SHORT		0b00001000
#define OPT_HELP_LONG		0b00010000
#define OPT_HELP		(OPT_HELP_SHORT | OPT_HELP_LONG)

enum OPT_HAS_ARG {
  OPT_NONE = 0,
  OPT_OPTIONAL,
  OPT_REQUIRED
};

enum OPT_ARG_TYPE {
  OPT_BOOL,
  OPT_INT,
  OPT_FLOAT,
  OPT_STRING,
  OPT_FILE,
  OPT_DIRECTORY
};

typedef struct option {
  char short_opt;	     /* The short option (if any) */
  const char *long_opt;	     /* The long option (if any) */
  enum OPT_HAS_ARG has_arg;
  enum OPT_ARG_TYPE type;    /* The type of the argument */
  int *found_opt;	     /* Whether or not we found this option */
  int *found_arg;	     /* Whether or not we found an argument */
  void *arg;		     /* Option argument (if any) */
  const char **argstr;	     /* The string form of the argument */
  /* The callback should return zero on success or nonzero on error, in
     which case the last argument (which must not be written beyond
     OPT_ERR_MAX, including the terminating null byte) must be set to a
     string describing the error. */
  union {
    int (*callback_none)(void *, char *);
    int (*callback_bool)(int, int, void *, char *);
    int (*callback_int)(int, long, void *, char *);
    int (*callback_float)(int, float, void *, char *);
    int (*callback_string)(int, const char *, void *, char *);
    int (*callback_file)(int, const char *, FILE *, void *, char *);
    int (*callback_directory)(int, const char *, DIR *, void *, char *);
  };
  void *data;		     /* Data to pass to that function */
  union {
    struct {
      int lower, upper;	     /* Lower and upper bounds for integers */
    } ibound;
    struct {
      double lower, upper;   /* Lower and upper bounds for floats */
    } fbound;
    const char *file_mode;   /* Mode to pass to fopen() */
  };
  const char *arg_help;	     /* Format of the argument (e.g. "FILE",
				"WxH", etc.) */
  const char *help;	     /* Description of the option */
} OPTION;

int parse_opts(int *, char **, const char *, const char *,
	       const char *, const OPTION *, int);
int format_help(FILE *file, const char *, const char *,
		const char *, const char *, const OPTION *);

#endif /* !_OPTIONS_H */
