/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2019  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * safe.c -- versions of standard functions that do not return on
 * error
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <errno.h>

#include "compat.h"
#include "format.h"
#include "safe.h"

#define SAFE_DIE(fmt, ...)					\
  do {								\
    fprintf(stderr, "%s:%d in function %s: " fmt ": %s\n",	\
	    file, line, func, __VA_ARGS__, strerror(errno));	\
    exit(-1);							\
  } while (0)

/* Call another safe function from within one. */
#define SAFE_CALL(name, ...)				\
  _SAFE_TR_NAME(name)(file, line, func, __VA_ARGS__)

void *_SAFE_DECL(xmalloc, size_t size) {
  void *ptr = malloc(size);

  if (ptr)
    return ptr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate %zu bytes of memory", size);
}

void *_SAFE_DECL(xcalloc, size_t nmemb, size_t size) {
  void *ptr = calloc(nmemb, size);

  if (ptr)
    return ptr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate %zu bytes of memory", size);
}

char *_SAFE_DECL(xstrdup, const char *str) {
  char *newstr = strdup(str);

  if (newstr)
    return newstr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate memory for a string of length %zu, \"%s\"",
	   strlen(str), str);
}

/*
 * These printf()-style functions return 'unsigned' simply to indicate
 * that the return value will never be negative. They will not,
 * however, return a value greater than INT_MAX.
 */

unsigned _SAFE_DECL(xasprintf, char **strp, const char *fmt, ...) {
  va_list ap;
  int ret;

  va_start(ap, fmt);
  ret = SAFE_CALL(xvasprintf, strp, fmt, ap);
  va_end(ap);

  return ret;
}

unsigned _SAFE_DECL(xvasprintf, char **strp, const char *fmt, va_list ap) {
  int ret = vasprintf(strp, fmt, ap);
  if (ret < 0)
    SAFE_DIE("failed to allocate for a string with format \"%s\"", fmt);
  return ret;
}

/*
 * These are safe versions of functions provided by Mu itself.
 */

void _SAFE_DECL(xformat,
		FILE *stream, unsigned short width, unsigned short *col,
		unsigned short indent, unsigned short subindent,
		const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  SAFE_CALL(xvformat, stream, width, col, indent, subindent, fmt, ap);
  va_end(ap);
}

char *_SAFE_DECL(xformat_string,
		 unsigned short width, unsigned short *col,
		 unsigned short indent, unsigned short subindent,
		 const char *fmt, ...) {
  va_list ap;
  char *ret;

  va_start(ap, fmt);
  ret = SAFE_CALL(xvformat_string, width, col, indent, subindent, fmt, ap);
  va_end(ap);

  return ret;
}

void _SAFE_DECL(xvformat,
		FILE *stream, unsigned short width, unsigned short *col,
		unsigned short indent, unsigned short subindent,
		const char *fmt, va_list ap) {
  if (vformat(stream, width, col, indent, subindent, fmt, ap))
    SAFE_DIE("failed to format a string with format \"%s\" to a file", fmt);
}

char *_SAFE_DECL(xvformat_string,
		 unsigned short width, unsigned short *col,
		 unsigned short indent, unsigned short subindent,
		 const char *fmt, va_list ap) {
  char *ret = vformat_string(width, col, indent, subindent, fmt, ap);
  if (!ret)
    SAFE_DIE("failed to format a string with format \"%s\" to a string", fmt);
  return ret;
}
